#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ----------------------------------------------------
# ChatGPT 翻译（国内需爬梯）
#   接口调用费用大约需要 $1/20万字
# ----------------------------------------------------


import os
import time
import openai
from ._settings import *
from ._trans_base import BaseTranslation

HTTP_PROXY = "HTTP_PROXY"
HTTPS_PROXY = "HTTPS_PROXY"

# GPT 接口模型定义 https://platform.openai.com/docs/models/
CHATGPT_35_TURBO = "gpt-3.5-turbo"
CHATGPT_4 = "gpt-4" # 8K
CHATGPT_4_TRUBO = "gpt-4-1106-preview"    # 128K
CHATGPT_4o = "gpt-4o"    # 128K

ARG_ROLE = 'role'
ARG_OPENAI_MODEL = 'openai_model'
ARG_PROXY_IP = 'proxy_ip'
ARG_PROXY_PORT = 'proxy_port'

class ChatgptTranslation(BaseTranslation) :

    RETRY = 3
    RETRY_WAIT_SECONDS = 30

    def __init__(self, openai_key, openai_model=CHATGPT_35_TURBO, proxy_ip='127.0.0.1', proxy_port=0) :
        BaseTranslation.__init__(self, '', openai_key)
        openai.api_key = openai_key
        self.model = openai_model or CHATGPT_35_TURBO
        self.proxy = f"http://{proxy_ip}:{proxy_port}" if proxy_port > 0 else ""
        
    
    def _translate(self, segment, from_lang='英文', to_lang='中文', args={}) :
        """
        翻译文本段落
        :param segment   : 待翻译文本段落
        :param from_lang : 待翻译文本的源语言（不同的平台语言代码不一样）
        :param to_lang   : 需要翻译的目标语言（不同的平台语言代码不一样）
        :param args      : hash 其他参数表
                            role : AI 角色人设
        :return: 已翻译的文本段落
        """
        role = args.get(ARG_ROLE) or f"您作为一个资深的专业翻译官，把下面文本中的{from_lang}内容翻译为{to_lang}，并润色（请勿回复翻译文本以外的内容）"
        role_setting = {"role": "system", "content": role}     # 设置 GPT 人设
        return self._ask_gpt(role_setting, segment)

    
    def _ask_gpt(self, role_setting, segment) :
        self._enable_proxy()
        msg = [
            role_setting, 
            {"role": "user", "content": segment}
        ]
        rsp = self._wait_for_ask(msg)
        rst = rsp.get("choices")[0]["message"]["content"]
        self._disable_proxy()
        return rst
    

    def _wait_for_ask(self, question) :
        rsp = ""
        for i in range(self.RETRY) :
            try :
                rsp = openai.ChatCompletion.create(
                    model=self.model,
                    messages=question
                )
                break

            except Exception as e:
                rsp = {
                    'choices': [{
                        'message': {
                            'content': f"[ERROR] ChatGPT No Response: {e}"
                        }
                    }]
                }
                time.sleep(self.RETRY_WAIT_SECONDS)
        return rsp


    def _enable_proxy(self) :
        os.environ[HTTP_PROXY] = self.proxy
        os.environ[HTTPS_PROXY] = self.proxy


    def _disable_proxy(self) :
        os.environ[HTTP_PROXY] = ""
        os.environ[HTTPS_PROXY] = ""

    
    def _len_limit(self) :
        return 500
    


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ----------------------------------------------------
# 腾讯机翻器（比谷歌准确，且每个月有 500 万字的免费额度）
# ----------------------------------------------------

from ._settings import *
from ._trans_base import BaseTranslation
from tencentcloud.common.credential import Credential
from tencentcloud.tmt.v20180321.tmt_client import TmtClient
from tencentcloud.tmt.v20180321.models import TextTranslateRequest
from tencentcloud.tmt.v20180321.models import TextTranslateResponse

GZ_REGION = 'ap-guangzhou'
ARG_REGION = 'region'
ARG_UNTRANSLATED_TEXT = 'UntranslatedText'

class TencentTranslation(BaseTranslation) :

    def __init__(self, api_id, api_key, region=GZ_REGION) :
        BaseTranslation.__init__(self, api_id, api_key)
        cred = Credential(api_id, api_key)
        self.client = TmtClient(cred, region)

    
    def _translate(self, segment, from_lang='en', to_lang='zh', args={}) :
        """
        翻译文本段落
        :param content      : 待翻译文本段落
        :param from_lang    : 待翻译文本的源语言（不同的平台语言代码不一样）
        :param to_lang      : 需要翻译的目标语言（不同的平台语言代码不一样）
        :param args         : hash 其他参数表
                                UntranslatedText: 忽略的翻译文本
        :return: 已翻译的文本段落
        """
        
        req = TextTranslateRequest()
        req.SourceText = segment
        req.Source = from_lang
        req.Target = to_lang
        req.ProjectId = 0
        req.UntranslatedText = args.get(ARG_UNTRANSLATED_TEXT) or ''   # 忽略的翻译文本
        rsp = self.client.TextTranslate(req)
        return rsp.TargetText


    def _len_limit(self) :
        return 2000
    
    



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------
# the inclusion of the tests module is not meant to offer best practices for
# testing in general, but rather to support the `find_packages` example in
# setup.py that excludes installing the "tests" package
# ----------------------------------------------------------------------
# 把父级目录（项目根目录）添加到工作路径，以便在终端也可以执行单元测试
import os
import sys
sys.path.append(os.path.dirname(os.path.abspath(__file__)) + '/../')
# ----------------------------------------------------------------------

import unittest
from src.transgpt.translate import *
from src.transgpt.trans_baidu import *
from src.transgpt.trans_tencent import *
from src.transgpt.trans_chatgpt import *


class TestScenes(unittest.TestCase):

    @classmethod
    def setUpClass(cls) :
        pass


    @classmethod
    def tearDownClass(cls) :
        pass


    def setUp(self) :
        # Baidu 生成 API Key https://fanyi-api.baidu.com/manage/developer
        self.BD_APP_ID = 'xxxxxxxxxxxxxxxx'
        self.BD_APP_KEY = 'yyyyyyyyyyyyyyyy'
        self.BD_SAVE_PATH = './test/output/result_baidu.txt'

        # Tencent 生成 API Key https://console.cloud.tencent.com/cam/capi
        self.TX_SECRET_ID = 'xxxxxxxxxxxxxxxx'
        self.TX_SECRET_KEY = 'yyyyyyyyyyyyyyyy'
        self.TX_SAVE_PATH = './test/output/result_tencent.txt'

        # ChatGPT 生成 API Key https://platform.openai.com/account/api-keys
        self.GPT_API_KEY = 'sk-xxxxxxxxxxxxxxxx'
        self.GPT_AI_ROLE = '基于《从零开始的异世界生活》小说的背景，把日文内容翻译成中文，并润色'
        self.GPT_SAVE_PATH = './test/output/result_chatgpt.txt'

        # 待翻译文本
        self.CONTENT = '''
「――村長くん、ちょっといいかい？」
そう背中から声をかけられ、ヴィンセント・ヴォラキアは足を止めた。
振り向く前から、声の相手はわかっている。一度聞いた声や、目にした相手のことは忘れない。これもすぐ、金髪に青い目をした行商人――フロップ・オコーネルとわかった。
「生憎、今は帝国存亡の危機だ。貴様の雑話にかまけている時間はない」
わかったが、わかった上でヴィンセントは取り合わない。
今しがた受け取った『星詠み』ウビルクの託宣、その扱いについてベルステツやセリーナたちと協議しなくてはならない。
敵への対抗策になると指名されたグルービーだが、死んでいる可能性が高い以上、それを当てにした計画を立てるなどと愚の骨頂だ。その点、指名されたもう一人の説得に向かったナツキ・スバルに期待する方が、まだマシな愚考と言えるだろう。
そもそも、あの少女を説得できる目はヴィンセントにはない。
彼女とはまともな意思疎通さえままならないのだ。ヴィンセントの持つ手札が少女に通用しない以上、二つの光の片方はスバルに掴み取らせるのが道理。
故に、ヴィンセントにできるのは現実的な策の検討の積み重ねだけ――、
「おっとっと、雑話かどうかは話してみなくちゃわからないことじゃないかな！僕がどんな話のタネを持っているか、いくら賢い村長くんでも水撒きする前からわかったりはしないだろうに！」
'''


    def tearDown(self) :
        pass


    # 使用 baidu 翻译（通用接口, 最全参数）
    def test_baidu_trans_1(self) :
        result = trans(
            self.CONTENT, 
            from_lang='jp', 
            to_lang='zh', 
            platform=BAIDU, 
            api_id=self.BD_APP_ID, 
            api_key=self.BD_APP_KEY, 
            savepath=self.BD_SAVE_PATH, 
            oncesave=True, 
        )


    # 使用 baidu 翻译（专用接口）
    def test_baidu_trans_2(self) :
        client = BaiduTranslation(
            api_id=self.BD_APP_ID, 
            api_key=self.BD_APP_KEY, 
            api_url=BAIDU_API_URL            # Option: 百度一般不会更换 API URL
        )

        # 最少参数（写文件）
        result = client.translate(
            self.CONTENT, 
            from_lang='jp',
            to_lang='zh',
            savepath=self.BD_SAVE_PATH, 
            oncesave=True
        )

        # 最少参数（仅内存）
        result = client.translate(
            self.CONTENT, 
            from_lang='jp',
            to_lang='zh'
        )


    # 使用 tencent 翻译（通用接口, 最全参数）
    def test_tencent_trans_1(self) :
        result = trans(
            self.CONTENT, 
            from_lang='ja', 
            to_lang='zh', 
            platform=TENCENT, 
            api_id=self.TX_SECRET_ID, 
            api_key=self.TX_SECRET_KEY, 
            savepath=self.TX_SAVE_PATH, 
            oncesave=True, 
            args={
                ARG_REGION: GZ_REGION, 
                ARG_UNTRANSLATED_TEXT: '『星詠み』'     # Option: 忽略不翻译的内容
            }
        )


    # 使用 tencent 翻译（专用接口）
    def test_tencent_trans_2(self) :
        client = TencentTranslation(
            api_id=self.TX_SECRET_ID, 
            api_key=self.TX_SECRET_KEY, 
            region=GZ_REGION             # Option: 选择最近自己位置的区域，可以加速接口网络
        )

        # 最少参数（写文件）
        result = client.translate(
            self.CONTENT, 
            from_lang='ja',
            to_lang='zh',
            savepath=self.TX_SAVE_PATH, 
            oncesave=True, 
            args={
                ARG_UNTRANSLATED_TEXT: '『星詠み』'     # Option: 忽略不翻译的内容
            }
        )

        # 最少参数（仅内存）
        result = client.translate(
            self.CONTENT, 
            from_lang='ja',
            to_lang='zh'
        )



    # 使用 ChatGPT 翻译（通用接口, 最全参数）
    def test_chatgpt_trans_1(self) :
        result = trans(
            self.CONTENT, 
            from_lang='日文',                   # Option: 通过指定 from_lang ，默认在内部生成 “翻译官” 的人设
            to_lang='中文',                     # Option: 通过指定 to_lang ，默认在内部生成 “翻译官” 的人设
            platform=CHATGPT, 
            api_id='', 
            api_key=self.GPT_API_KEY, 
            savepath=self.GPT_SAVE_PATH, 
            oncesave=False, 
            args={
                ARG_ROLE: self.GPT_AI_ROLE,     # Option: 直接定义人设，无需指定 from_lang 和 to_lang （内部生成的默认人设会被覆盖）
                ARG_OPENAI_MODEL: CHATGPT_35_TURBO, 
                ARG_PROXY_IP: '127.0.0.1', 
                ARG_PROXY_PORT: 8888
            }
        )


    # 使用 ChatGPT 翻译（专用接口）
    def test_chatgpt_trans_2(self) :
        client = ChatgptTranslation(
            self.GPT_API_KEY,
            openai_model=CHATGPT_35_TURBO,  # Option: 使用的模型，直接决定调用费用
            proxy_ip='127.0.0.1',           # Option: 国内需要挂代理
            proxy_port=8888                 # Option: 国内需要挂代理
        )

        # 最全参数
        result = client.translate(
            self.CONTENT, 
            savepath=self.GPT_SAVE_PATH, 
            oncesave=False, 
            from_lang='日文',                   # Option: 通过指定 from_lang ，默认在内部生成 “翻译官” 的人设
            to_lang='中文',                     # Option: 通过指定 to_lang ，默认在内部生成 “翻译官” 的人设
            args={
                ARG_ROLE: self.GPT_AI_ROLE,    # Option: 直接定义人设，无需指定 from_lang 和 to_lang （内部生成的默认人设会被覆盖）
            }
        )

        # 最少参数（写文件，默认人设）
        result = client.translate(
            self.CONTENT, 
            from_lang='日文',
            to_lang='中文',
            savepath=self.GPT_SAVE_PATH
        )

        # 最少参数（写文件，自定义人设）
        result = client.translate(
            self.CONTENT, 
            savepath=self.GPT_SAVE_PATH, 
            args={
                ARG_ROLE: self.GPT_AI_ROLE,
            }
        )

        # 最少参数（仅内存，默认人设）
        result = client.translate(
            self.CONTENT, 
            from_lang='日文',
            to_lang='中文'
        )

        # 最少参数（仅内存，自定义人设）
        result = client.translate(
            self.CONTENT, 
            args={
                ARG_ROLE: self.GPT_AI_ROLE,
            }
        )
        

if __name__ == '__main__':
    unittest.main()


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# --------------------------------------------
# 文本翻译通用接口
# --------------------------------------------

from ._settings import *
from .trans_baidu import *
from .trans_tencent import *
from .trans_chatgpt import *
from color_log.clog import log


# 翻译平台
BAIDU = "baidu"
TENCENT = "tencent"
CHATGPT = "chatgpt"


def trans(content, api_id, api_key, 
            from_lang='', to_lang='', 
            platform=TENCENT, savepath='', oncesave=False, 
            args={}
    ) -> str:
    """
    翻译文本
    :param content      : 待翻译文本
    :param api_id       : 接口 API_ID / APP_ID （ChatGPT 不需要）
    :param api_key      : 接口 API_KEY / SECRET_KEY
    :param from_lang    : 待翻译文本的源语言（不同的平台语言代码不一样）
    :param to_lang      : 需要翻译的目标语言（不同的平台语言代码不一样）
    :param platform     : 翻译平台，目前支持： chatgpt, baidu, tencent（默认）
    :param savepath     : 翻译文本保存的位置，若为空则不保存（可通过返回值获取）
    :param oncesave     : 是否一次性保存翻译文本（对于超长文本，内部会进行分段翻译，为了避免网络异常导致已翻译文本丢失，此项默认关闭）
    :param args         : hash 其他参数表
                            api_url         : 仅百度翻译有用：接口 API 路径
                            region          : 仅腾讯翻译有用：翻译服务器所在的区域
                            UntranslatedText: 仅腾讯翻译有用：忽略的翻译文本
                            role            : 仅 ChatGPT 有用：AI 角色人设
                            openai_model    : 仅 ChatGPT 有用：OpenAPI 接口的大模型
                            proxy_ip        : 仅 ChatGPT 有用：网络代理服务 IP
                            proxy_port      : 仅 ChatGPT 有用：网络代理服务端口
    :return: 已翻译文本
    """

    log.info(f"正在使用 [{platform}] 翻译文本 ...")
    if platform == BAIDU :
        api_url = args.get(ARG_API_URL) or BAIDU_API_URL
        client = BaiduTranslation(api_id, api_key, api_url)

    elif platform == CHATGPT :
        openai_model = args.get(ARG_OPENAI_MODEL) or CHATGPT_35_TURBO
        proxy_ip = args.get(ARG_PROXY_IP) or '127.0.0.1'
        proxy_port = args.get(ARG_PROXY_PORT) or 0
        client = ChatgptTranslation(api_key, openai_model, proxy_ip, proxy_port)
    
    else :
        region = args.get(ARG_REGION) or GZ_REGION
        client = TencentTranslation(api_id, api_key, region)

    trans_content = client.translate(content, from_lang, to_lang, savepath, oncesave, args)
    log.info(f"文本翻译完成")
    if savepath :
        log.info("翻译文本已%s存储到: %s" % ('' if oncesave else '分段', savepath))
    return trans_content
    



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ----------------------------------------------------

import os
import time
from ._settings import *
from abc import ABCMeta, abstractmethod     # python不存在抽象类的概念， 需要引入abc模块实现
from color_log.clog import log


class BaseTranslation :

    __metaclass__ = ABCMeta # 定义为抽象类

    def __init__(self, api_id, api_key) -> None :
        """
        翻译文本
        :param api_id  : 接口 API_ID / APP_ID （ChatGPT 不需要）
        :param api_key : 接口 API_KEY / SECRET_KEY
        """
        self.api_id = api_id
        self.api_key = api_key


    def translate(self, content, from_lang='', to_lang='', savepath='', oncesave=False, args={}) -> str :
        """
        翻译文本
        :param content   : 待翻译文本
        :param from_lang : 待翻译文本的源语言（不同的平台语言代码不一样）
        :param to_lang   : 需要翻译的目标语言（不同的平台语言代码不一样）
        :param savepath  : 翻译文本保存的位置，若为空则不保存（可通过返回值获取）
        :param oncesave  : 是否一次性保存翻译文本（对于超长文本，内部会进行分段翻译，为了避免网络异常导致已翻译文本丢失，此项默认关闭）
        :param args      : hash 其他参数表
                            api_url         : 仅百度翻译有用：接口 API 路径
                            region          : 仅腾讯翻译有用：翻译服务器所在的区域
                            UntranslatedText: 仅腾讯翻译有用：忽略的翻译文本
                            role            : 仅 ChatGPT 有用：AI 角色人设
                            openai_model    : 仅 ChatGPT 有用：OpenAPI 接口的大模型
                            proxy_ip        : 仅 ChatGPT 有用：网络代理服务 IP
                            proxy_port      : 仅 ChatGPT 有用：网络代理服务端口
        :return: 已翻译文本
        """
        
        self.oncesave = oncesave    # False: 分段保存； True: 一次性保存
        trans_result = []
        segments = self._cut(content)
        log.info("切割为 [%i] 段翻译 ..." % len(segments))

        cnt = 0
        for seg in segments :
            cnt += 1
            log.info("正在翻译第 [%i] 段 ..." % cnt)
            trans_segment = self._translate(seg, from_lang, to_lang, args)
            trans_result.append(trans_segment)
            if not self.oncesave :
                self._save_trans(trans_segment, savepath)
            time.sleep(1)

        trans_content = DOUBLE_CRLF.join(trans_result)
        if self.oncesave :
            self._save_trans(trans_content, savepath)
        return trans_content


    # 翻译接口每次调用有字数限制，过长文本需要分段翻译。
    def _cut(self, content) :
        segments = []
        seg = []
        seg_len = 0

        # 为了避免语义问题，此处会使用换行符对文本进行多段切割，
        lines = content.split(CRLF)
        for line in lines :
            line = line.strip()
            if not line :
                continue

            # 然后按顺序重新组装为“不大于限制字数”的段落
            line_len = len(line)
            if seg_len + line_len > self._len_limit() :
                segments.append(CRLF.join(seg))
                seg_len = 0
                seg = []

            seg.append(line)
            seg_len += line_len

        segments.append(DOUBLE_CRLF.join(seg))
        return segments
    

    def _save_trans(self, content, savepath) :
        if not savepath :
            return
        
        save_dir = os.path.dirname(savepath)
        try :
            os.makedirs(save_dir)
        except :
            pass

        if self.oncesave :
            with open(savepath, "w+", encoding=CHARSET) as file :
                file.write(content)
            log.info(f"翻译完成，译文已存储到 [{savepath}]")

        else :
            with open(savepath, "a+", encoding=CHARSET) as file :
                file.write(content)
            log.info(f"段翻译完成，译文已追加存储到 [{savepath}]")


    
    @abstractmethod
    def _len_limit(self) :
        return 500  # API 接口限制每次翻译的字数
    

    @abstractmethod
    def _translate(self, segment, from_lang='en', to_lang='zh', args={}) -> str :
        # TODO 子类实现
        return ""




#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ----------------------------------------------------
# 百度机翻器（比谷歌准确，且每个月有 200 万字的免费额度）
# ----------------------------------------------------

import time
import hashlib
import requests
import json
from ._settings import *
from ._trans_base import BaseTranslation
from color_log.clog import log


BAIDU_API_URL = "https://fanyi-api.baidu.com/api/trans/vip/translate"
ARG_API_URL = 'api_url'

class BaiduTranslation(BaseTranslation) :

    def __init__(self, api_id, api_key, api_url=BAIDU_API_URL) -> None :
        BaseTranslation.__init__(self, api_id, api_key)
        self.api_url = api_url


    def _translate(self, segment, from_lang='en', to_lang='zh', args={}) :
        """
        翻译文本段落
        :param segment   : 待翻译文本段落
        :param from_lang : 待翻译文本的源语言（不同的平台语言代码不一样）
        :param to_lang   : 需要翻译的目标语言（不同的平台语言代码不一样）
        :param args      : hash 其他参数表（暂时没用）
        :return: 已翻译的文本段落
        """

        salt, sign = self._to_sign(segment)
        headers = {
            'Content-Type': 'application/x-www-form-urlencoded'
        }
        body = {
            'q': segment, 
            'from': from_lang, 
            'to': to_lang, 
            'appid': self.api_id, 
            'salt': salt, 
            'sign': sign
        }
        
        response = requests.post(self.api_url, headers=headers, data=body)
        trans_result = []
        try :
            if response.status_code == 200:
                rst = json.loads(response.text)
                for line in rst.get("trans_result") :
                    trans_result.append(line.get("dst").strip())
            else :
                log.error(f"翻译段落失败: 接口 [{response.status_code}] 异常")
        except :
            log.error(f"翻译段落失败: {response.text}")
        return DOUBLE_CRLF.join(trans_result)


    def _to_sign(self, data) :
        salt = int(time.time())
        sign = hashlib.md5(
            ("%s%s%i%s" % (
                self.api_id, 
                data,
                salt, 
                self.api_key
            )).encode(encoding=CHARSET)
        ).hexdigest()
        return salt, sign


    def _len_limit(self) :
        return 2000
    